import inspect
import re
from typing import Dict
from typing import List


def ignore_parameter_changes(
    hub,
    ignore_changes: List,
    params: Dict,
    param_signatures: Dict[str, inspect.Parameter],
):
    """
    Set an optional parameter within params to None according to the parameter path in ignore_changes. For example,
    path key1:key2 path will result to {key1: {key2: None}}
    :param ignore_changes: A list of path of parameters that will be assigned to None value
                           to ignore being updated in present().
    :param params: A dict of parameter-value pairs.
    :param param_signatures: A dict containing the parameter signatures.
    """
    for param_path in ignore_changes:
        # Split the path with ":"
        path = param_path.split(":")
        arg_key, indexes = _parse_index(path[0])
        if arg_key in ["hub", "ctx", "name"]:
            continue
        # We can only assigning None to the optional parameters which default is not inspect._empty
        if (
            arg_key in param_signatures
            and param_signatures.get(arg_key).default != inspect._empty
        ):
            try:
                _nullify_parameter(parent=params, indexes=[], remaining_keys=path)
            except Exception as e:
                hub.log.warning(
                    f"Error when processing ignore_changes parameter path {param_path}: {e.__class__.__name__} {e}"
                )


def _parse_index(key_to_parse):
    """
    Parse indexes of key. For example, test[0][1] will return "test" as parsed key and [0,1] as parsed indexes.
    test[0][*] will return 'test' as parsed key and [0, '*'] as the parsed indexes.
    """
    rule = r"\[\d+\]|\[\*\]"
    indexes = re.findall(rule, key_to_parse)
    if indexes:
        index_digits = []
        for index in indexes:
            if "[*]" == index:
                index_digits.append("*")
            else:
                index_digit = re.search(r"\d+", index).group(0)
                index_digits.append(int(index_digit))

        return key_to_parse[0 : key_to_parse.index("[")], index_digits

    return key_to_parse, None


def _nullify_parameter(parent: Dict or List, indexes: List, remaining_keys: List):
    """
    Go through a dictionary or list according to the path and replace the destination value with None.
    For example, if path is a:b, params {"a": {"b": "value"}} will become {"a": {"b": None}}
    :param parent: A Dict or List collection that contains the data to be set to None
    :param indexes: A list that contains the sequence of indexes to be traversed if the parent is a list
    :param remaining_keys: a List of remaining keys that need to be traversed.
    """
    # If there is no remaining keys, then we are in the last part of a parameter path
    if (not remaining_keys) and indexes:
        # If indexes is not empty, parent should be a List type collection, which allows us to traverse it with the indexes
        if len(indexes) == 1:
            # If indexes only has one index left, then we we can just replace the value at the index with None
            if isinstance(indexes[0], int):
                parent[indexes[0]] = None
            else:
                raise ValueError(f"Invalid index {indexes[0]} at the end.")
        else:
            # Traverse the param data according to the indexes
            if isinstance(indexes[0], int):
                _nullify_parameter(parent[indexes[0]], indexes[1:], [])
            # If '*' is in the index, then we traverse into all elements of the list
            elif indexes[0] == "*":
                for element in parent:
                    _nullify_parameter(element, indexes[1:], [])
            else:
                raise ValueError(f"Invalid index {indexes[0]}")
    elif indexes:
        # If indexes exists, parent should be a List type collection. We traverse it with the indexes
        if isinstance(indexes[0], int):
            _nullify_parameter(
                parent[indexes[0]],
                indexes[1:] if len(indexes) > 1 else [],
                remaining_keys,
            )
        elif indexes[0] == "*":
            for element in parent:
                _nullify_parameter(
                    element, indexes[1:] if len(indexes) > 1 else [], remaining_keys
                )
        else:
            raise ValueError(f"Invalid index {indexes[0]}")
    else:
        # If indexes is empty, we can move forward to process the next element in remaining_keys, which holds the
        # remaining parameter path
        arg_key, indexes = _parse_index(remaining_keys[0])
        if len(remaining_keys) == 1 and not indexes:
            # If the remaining key is one, and there is no index in the key, then this is the last key in the data
            # so we can just override it with None
            parent[arg_key] = None
        else:
            _nullify_parameter(
                parent[arg_key],
                indexes,
                remaining_keys[1:] if len(remaining_keys) else [],
            )



from typing import Any
from typing import Dict

__func_alias__ = {"ctx_": "ctx"}
__contracts__ = ["returns", "soft_fail"]


def __init__(hub):
    hub.exec.test.ITEMS = {}


def ping(hub):
    """
    Immediately return success
    """
    return {"result": True, "ret": True}


async def aping(hub):
    return {"result": True, "ret": True}


def fail(hub):
    raise Exception("Expected failure")


async def afail(hub):
    raise Exception("Expected failure")


def echo(hub, value):
    """
    Return the parameter passed in without changing it at all
    """
    return value


async def aecho(hub, value):
    """
    Return the parameter passed in without changing it at all
    """
    return value


async def event(
    hub,
    body: str,
    ingress_profile: str = "default",
    tags: Dict[str, Any] = None,
):
    """
    :param body: The event body
    :param ingress_profile: The ingress profile to allowlist for this event

    .. code-block:: bash

        $ idem exec test.event body ingress_profile="default" --serialize-plugin="json"
        $ idem exec test.event body="my_event" ingress_profile="default" --serialize-plugin="json"
    """
    event_kwargs = dict(body=body, profile=ingress_profile, tags=tags)
    await hub.idem.event.put(**event_kwargs)
    return {"result": True, "ret": event_kwargs}


async def reconcile_with_rerun_data(hub, ctx):
    """
    In order to trigger reconciliation for exec.run the result should be False.
    Also 'rerun_data' passed through the context will be concatenated.
    """
    return {
        "result": False,
        "rerun_data": ctx.get("rerun_data") + "1-" if ctx.get("rerun_data") else "-1-",
        "comment": "test",
        "ret": {"hello": "world"},
    }



def stage(hub, name: str):
    """
    Apply the exclude value
    """
    low = hub.idem.ccomps.treq.apply(
        hub.idem.RUNS[name]["subs"],
        hub.idem.RUNS[name]["low"],
    )
    hub.idem.RUNS[name]["low"] = low



from typing import Any
from typing import Dict
from typing import List


def check(hub, rlist: List[Dict[str, Any]]) -> List[str]:
    """
    Check if any of the required list passes. If a single check passes then
    all is well, only return the errors is ALL FAIL
    """
    errors = []
    good = False
    for rdat in rlist:
        err = rdat.get("errors", [])
        if not err:
            good = True
        errors.extend(rdat.get("errors", []))
    if good:
        return []
    return errors



from typing import Iterable


def sig_create(hub, iterable: Iterable, **kwargs) -> Iterable:
    """
    Create a progress bar
    """


def sig_update(hub, progress_bar: Iterable, **kwargs):
    """
    Update the progress bar from the creation function
    """



import asyncio
import copy
from typing import Any
from typing import Dict
from typing import List

import dict_tools.differ as differ

from idem.idem.state import Status

# Reconciliation loop stops after MAX_RERUNS_WO_CHANGE reruns without change.
# This is to make sure we do not retry forever on failures that cannot be fixed by
# reconciliation.
MAX_RERUNS_WO_CHANGE = 3
# Sleep time in seconds between re-runs
DEFAULT_RECONCILE_WAIT = 3
# Dictionary for the default static sleep time
DEFAULT_STATIC_RECONCILE_WAIT = {"static": {"wait_in_seconds": DEFAULT_RECONCILE_WAIT}}
# Dictionary keeping state's reconcile wait time in seconds
_state_to_sleep_map = {}


async def loop(
    hub,
    pending_plugin: str,
    name: str,
    apply_kwargs: Dict[str, Any],
):
    """
    This loop attempts to apply states.
    This function returns once all the states are successful or after MAX_RERUNS_WO_CHANGE, whichever occur first.
    The sleep time between each attempt will be determined by a "wait" plugin and might change between each iterations.
    Reconciliation is required if the state is "pending" as defined by the pending plugin.
    The default pending plugin defines pending state if result  is not 'True' or there are 'changes'.

    @param hub:
    @param pending_plugin: plugin name for checking if the state is pending based on the result
    @param name: name of the run
    @param apply_kwargs: possible addition arguments
    :return: dictionary { "re_runs_count": <number of re-runs that occurred>,
                "require_re_run": <True/False whether the last run require more reconciliation> }
    """
    # Record the first run as following reconciliation re-runs will
    # include subset of pending states
    first_run = copy.deepcopy(hub.idem.RUNS[name]["running"])

    pending_tags = hub.reconcile.basic.get_pending_tags(pending_plugin, first_run)
    if not pending_tags:
        hub.idem.state.update_status(name, Status.FINISHED)
        return {"re_runs_count": 0, "require_re_run": False}

    # Populate wait time algorithm and values for the different states
    # in this run. State has to define __reconciliation_wait__
    # with values such as:
    # { "exponential": {"wait_in_seconds": 2, "multiplier": 10} }
    # { "static": {"wait_in_seconds": 3} }
    # { "random": {"min_value": 1, "max_value": 10} }
    hub.reconcile.basic.populate_wait_times(first_run)

    # Populate old_state for pending tags
    tag_to_old_state = _populate_old_states(first_run)
    # Concatenate comments for pending tags
    tag_to_comments = hub.reconcile.basic.populate_comments(first_run)

    count = 0
    count_wo_change = 0
    while count_wo_change < MAX_RERUNS_WO_CHANGE:
        sleep_time_sec = get_max_wait_time(hub, count, pending_tags)
        hub.log.debug(f"Sleeping {sleep_time_sec} seconds for {name}")
        await asyncio.sleep(sleep_time_sec)

        count = count + 1
        hub.log.debug(f"Retry {count} for {name}")

        last_run = hub.idem.RUNS[name]["running"]

        # Re-run pending states
        await hub.idem.run.init.start(name, pending_tags)

        # This run will include only the pending tags
        current_run = hub.idem.RUNS[name]["running"]
        tag_to_comments = hub.reconcile.basic.populate_comments(
            current_run, tag_to_comments
        )
        pending_tags = hub.reconcile.basic.get_pending_tags(pending_plugin, current_run)
        if not pending_tags:
            hub.reconcile.basic.update_result(
                name, first_run, current_run, tag_to_old_state, tag_to_comments
            )
            return {"re_runs_count": count, "require_re_run": False}

        if _is_same_result(hub, last_run, current_run):
            count_wo_change = count_wo_change + 1
        else:
            # reset the count w/o changes upon a change
            count_wo_change = 0

    hub.log.debug(
        f"Reconciliation loop returns after {count} runs total, and {count_wo_change} runs without any change."
    )

    hub.reconcile.basic.update_result(
        name, first_run, current_run, tag_to_old_state, tag_to_comments
    )
    return {
        "re_runs_count": count,
        "require_re_run": True,
    }


def get_pending_tags(hub, pending_plugin, runs):
    # invoke pending plugin and populate pending tags
    pending_tags = []
    for tag in runs:
        if hub.reconcile.pending[pending_plugin].is_pending(
            ret=runs[tag], state=_tag_2_state(tag)
        ):
            pending_tags.append(tag)

    return pending_tags


def populate_wait_times(hub, runs):
    # Populate sleep times per state
    for tag in runs:
        state = _tag_2_state(tag)
        if state not in _state_to_sleep_map.keys():
            try:
                _state_to_sleep_map[state] = getattr(
                    hub.states[state],
                    "__reconcile_wait__",
                    DEFAULT_STATIC_RECONCILE_WAIT,
                )
            except Exception as e:
                hub.log.error(
                    f"Failed to retrieve sleep time for state {state}: {e.__class__.__name__}: {e}"
                )
                _state_to_sleep_map[state] = DEFAULT_STATIC_RECONCILE_WAIT


def get_max_wait_time(hub, run_count, pending_tags: List[str]):
    # Return the maximum wait time among the pending tags
    max_sleep_time = DEFAULT_RECONCILE_WAIT
    for tag in pending_tags:
        state_wait = _state_to_sleep_map.get(
            _tag_2_state(tag), DEFAULT_STATIC_RECONCILE_WAIT
        )
        wait_alg = list(state_wait.keys())[0]
        wait_val = state_wait[wait_alg]
        sleep_time = hub.reconcile.wait[wait_alg].get(**wait_val, run_count=run_count)
        if sleep_time > max_sleep_time:
            max_sleep_time = sleep_time

    return max_sleep_time


def update_result(
    hub,
    name,
    first_run,
    last_run,
    tag_to_old_state: Dict[str, dict],
    tag_to_comments: Dict[str, tuple],
):
    # Merge last_run results into first_run, recalculate changes
    # Update 'old_state' from the original run and recalculate 'changes' to reflect
    # all the changes that occurred during reconciliation:
    # the delta between original run's old_state and last_run 'new_state'
    # Update 'comment' with the cumulative comments from all runs (w/o duplicates)
    # Set that as the 'running' on the hub
    if not last_run:
        return

    first_run.update(last_run)
    for tag in last_run:
        orig_old_state = tag_to_old_state.get(tag, None)
        last_old_state = last_run[tag].get("old_state", None)
        if orig_old_state != last_old_state:
            hub.log.debug(
                f"Replacing 'old_state' for '{_tag_2_state(tag)}': {last_old_state} with {orig_old_state}"
            )

            first_run[tag]["old_state"] = orig_old_state

            # Only when replacing 'old_state' 'changes' have to be re-calculated
            first_run[tag]["changes"] = differ.deep_diff(
                last_run[tag].get("old_state") or {},
                last_run[tag].get("new_state") or {},
            )
        if tag_to_comments.get(tag):
            first_run[tag]["comment"] = tag_to_comments.get(tag)

    hub.idem.RUNS[name]["running"] = first_run


def populate_comments(
    hub, run, tag_to_comments: Dict[str, tuple] = None
) -> Dict[str, tuple]:
    # If tag_to_comments is empty populate it with the run comments.
    # Otherwise add comments to existing
    if not tag_to_comments:
        tag_to_comments = {}
    for tag in run:
        # We expect comments to be of type tuple
        if run[tag].get("comment"):
            comment = run[tag].get("comment")
            if isinstance(comment, str):
                hub.log.debug(
                    f"Expecting a comment of type tuple but got a str: {comment}"
                )
                comment = (comment,)
            elif isinstance(comment, list):
                hub.log.debug(
                    f"Expecting a comment of type tuple but got a list: {comment}"
                )
                comment = tuple(comment)
            elif not isinstance(comment, tuple):
                hub.log.warning(
                    f"Unsupported comment type: {type(comment)}, comment: {str(comment)}"
                )
                continue

            if tag_to_comments.get(tag) and not comment[0] in tag_to_comments.get(tag):
                tag_to_comments[tag] += comment
            else:
                tag_to_comments[tag] = comment

    return tag_to_comments


def _is_same_result(hub, last_run, current_run):
    for tag in current_run:
        if (
            last_run.get(tag)
            and last_run[tag]["result"] == current_run[tag]["result"]
            and last_run[tag]["changes"] == current_run[tag]["changes"]
        ):
            continue
        else:
            result1 = last_run[tag]["result"]
            result2 = current_run[tag]["result"]
            hub.log.debug(
                f"Changes between runs: last result: {result1} - current result: {result2}"
            )
            changes1 = last_run[tag]["changes"]
            changes2 = current_run[tag]["changes"]
            hub.log.debug(
                f"Changes between runs: last run changes: {changes1} - current run changes: {changes2}"
            )
            return False
    return True


def _populate_old_states(run):
    # Keep old_state per tag from the original run
    tag_to_old_state = {}
    for tag in run:
        if run[tag].get("old_state", None):
            tag_to_old_state[tag] = copy.deepcopy(run[tag]["old_state"])
    return tag_to_old_state


def _tag_2_state(tag):
    # if state is sls we need to remove the sls.run prefix to determine the exact state
    # we need exact state being run to populate timeouts and to call is_pending function of that state.
    state = tag[0 : tag.find("_|")]
    if state == "sls":
        tag_arr = tag.split("_|-")
        if len(tag_arr) > 4:
            tag = "_|-".join(tag_arr[2:])

    return tag[0 : tag.find("_|")]



from typing import Iterable

try:
    from tqdm.auto import tqdm

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def create(hub, iterable: Iterable, **kwargs) -> Iterable:
    """
    Create a tqdm progress bar that updates as the iterable is iterated
    """
    progress_bar = tqdm(
        iterable,
        **kwargs,
    )

    return progress_bar


def update(hub, progress_bar: Iterable, **kwargs):
    if not isinstance(progress_bar, tqdm):
        return
    progress_bar.update(**kwargs)



import pathlib


def gather(hub, profiles):
    """
    The "default" profile for esm.local is created with configuration values
    Other profiles can be created manually with the following format.

    .. code-block:: sls

        esm.local:
            profile_name:
                run_name: ...
                cache_dir: /path/to/cache/dir
                serial_plugin: json|msgpack
    """
    sub_profiles = {}
    for profile, ctx in profiles.get("esm.local", {}).items():
        idem_opts = hub.OPT.get("idem", {})
        # Fill in any missing params from hub.OPT if possible
        if not ctx.get("run_name"):
            ctx["run_name"] = idem_opts.get("run_name")
        if not ctx.get("cache_dir"):
            ctx["cache_dir"] = idem_opts.get("cache_dir")
        if not isinstance(ctx["cache_dir"], pathlib.Path):
            ctx["cache_dir"] = pathlib.Path(ctx["cache_dir"])
        if not ctx.get("serial_plugin"):
            ctx["serial_plugin"] = idem_opts.get("serial_plugin")
        sub_profiles[profile] = ctx
    return sub_profiles



import datetime
from typing import Any
from typing import Dict


def get_chunks_from_esm(hub, resource_type, name, sls_run_id=None):
    rets = []
    run_name = hub.idem.RUN_NAME
    # Get the current state
    esm_state: Dict[str, Any] = dict(hub.idem.RUNS[run_name]["managed_state"])

    for resource_in_state in esm_state:
        chunk = hub.tool.idem.esm.convert_state_data_to_chunk(
            esm_state[resource_in_state], resource_type, name, sls_run_id
        )
        esm_tag = hub.idem.tools.gen_chunk_esm_tag(chunk)
        if resource_in_state == esm_tag:
            chunk["resource_state"] = esm_state[resource_in_state]
            rets.append(chunk)
    return rets


def convert_state_data_to_chunk(
    hub, state_data: Dict, resource_type, name, sls_run_id=None
):
    if not isinstance(state_data, dict):
        hub.log.warning(
            f"ESM did not get a dictionary for state_data '{type(state_data)}'. Resource type: {resource_type}"
        )
        chunk = {}
    else:
        chunk = state_data
    chunk["state"] = resource_type
    chunk["__id__"] = name
    chunk["fun"] = "present"
    if sls_run_id:
        chunk["sls_run_id"] = sls_run_id
        chunk["__id__"] = sls_run_id + ":" + name
    if "name" not in chunk:
        chunk["name"] = name

    return chunk


def update_running_from_esm(hub, chunk):
    tag = hub.idem.tools.gen_chunk_func_tag(chunk)
    esm_tag = hub.idem.tools.gen_chunk_esm_tag(chunk)
    start_time = datetime.datetime.now()
    run_num = hub.idem.RUNS[hub.idem.RUN_NAME]["run_num"]
    return {
        "tag": tag,
        "name": chunk["name"],
        "__id__": chunk["__id__"],
        "changes": {},
        "new_state": chunk["resource_state"],
        "old_state": chunk["resource_state"],
        "comment": (),
        "result": True,
        "esm_tag": esm_tag,
        "__run_num": run_num,
        "start_time": str(start_time),
        "total_seconds": 0,
    }



from typing import Any
from typing import Dict


def run(
    hub,
    seq: Dict[int, Dict[str, Any]],
    low: Dict[str, Any],
    running: Dict[str, Any],
    options: Dict[str, Any],
) -> Dict[int, Dict[str, Any]]:
    """
    Process the recreate_on_update requisite.
    """
    for ind, data in seq.items():
        if "recreate_on_update" not in data["chunk"]:
            continue
        chunk = data["chunk"]
        r_tag = hub.idem.tools.gen_chunk_func_tag(chunk)
        reqret = {
            "r_tag": r_tag,
            "req": "recreate_on_update",
            "ret": {},
        }
        data["reqrets"].append(reqret)

    return seq



"""
This module is used to retrieve files from a local source
"""
import mimetypes
import os
import pathlib
from typing import ByteString
from typing import Tuple

__virtualname__ = "file"


async def process(
    hub, ctx, protocol: str, source_path: str, loc: str
) -> Tuple[str, ByteString]:
    """
    Determine if a file has mimetypes, if so, unravel it and save it's tree in memory
    """
    filemimetype, encoding = mimetypes.guess_type(source_path)
    if not filemimetype:
        file_extension = pathlib.Path(source_path).suffix
        if file_extension:
            filemimetype = file_extension

    if not filemimetype:
        return "", b""

    for plugin in hub.source:
        if not hasattr(plugin, "MIMETYPES"):
            continue
        if filemimetype in plugin.MIMETYPES:
            return await plugin.cache(
                ctx=ctx, protocol=protocol, source=source_path, location=loc
            )
    return "", b""


async def cache(
    hub, ctx, protocol: str, source: str, location: str
) -> Tuple[str, ByteString]:
    """
    Take a file from a location definition and cache it in memory
    """
    # Check if the file has a mimetype and call the right plugin
    source_path = source

    # If the source_path is a file then check if it is a mimetype, the location might be inside a compressed directory
    if os.path.isfile(source_path):
        ref, mime_source = await hub.source.file.process(
            ctx, protocol, source_path, location
        )
        if mime_source:
            return ref, mime_source

    # No mimetypes, read a plain SLS file
    full = os.path.join(source_path, location)
    if os.path.isfile(full):
        with open(full, "rb") as rfh:
            in_memory_file = rfh.read()

        # Use full file path as the unique sls_ref
        return full, in_memory_file



import enum
import json
import uuid
from typing import Any
from typing import Dict
from typing import List

import dict_tools.update

__func_alias__ = {"compile_": "compile"}


class Status(enum.IntEnum):
    FINISHED = 0
    CREATED = 1
    GATHERING = 2
    COMPILING = 3
    RUNNING = 4
    COMPILATION_ERROR = -1
    GATHER_ERROR = -2
    RUNTIME_ERROR = -3
    UNDEFINED = -4


def update_status(hub, name: str, status_: Status):
    """
    Update the status of the run and fire an event
    """
    if name in hub.idem.RUNS:
        hub.idem.RUNS[name]["status"] = status_
    hub.idem.event.put_nowait(
        profile="idem-status",
        body=status_.name,
        tags={"ref": "idem.state.update_status", "type": "state-status"},
    )


async def create(
    hub,
    name: str,
    sls_sources: List[str],
    render: str,
    runtime: str,
    subs: List[str],
    cache_dir: str,
    test: bool,
    acct_file: str,
    acct_key: str,
    acct_profile: str,
    acct_blob: bytes,
    managed_state: Dict[str, Any],
    param_sources: List[str],
    acct_data: Dict[str, Any] = None,
    invert_state: bool = False,
):
    """
    Create a new instance to execute against
    """
    if acct_data is None:
        acct_data = await hub.idem.acct.data(
            acct_key=acct_key, acct_file=acct_file, acct_blob=acct_blob
        )

    hub.idem.RUNS[name] = {
        "sls_sources": sls_sources,
        "param_sources": param_sources,
        "render": render,
        "runtime": runtime,
        "subs": subs,
        "cache_dir": cache_dir,
        "states": {},
        "test": test,
        "resolved": set(),
        "files": set(),
        "high": {},
        # Jinja render uses the params python dict object to read the value corresponding to
        # any parameter. When the parameter is not present params dict object's get() returns
        # the None object. We need to return __PYTHON_NONE__ instead, so that it gets suitably
        # converted into the python None object in the replacement renderer.
        # Here we are using type library function to create a new type object, that has get()
        # function which returns self[key] if parameter (key) is present, default value (if
        # specified), and __PYTHON_NONE__ otherwise. Python dict is used as base class.
        # Thus, for all parameters that are not present in dict, this get() lambda returns
        # __PYTHON_NONE__.
        "params": type(
            "",
            (dict,),
            {
                "get": lambda self, key, value=None: self[key]
                if key in self
                else value
                if value is not None
                else "__PYTHON_NONE__"
            },
        )(),
        "post_low": [],
        "errors": [],
        "iorder": 100000,
        "sls_refs": {},
        "blocks": {},
        "running": {},
        "meta": {"SLS": {}, "ID_DECS": {}},
        "run_num": 1,
        "add_low": [],
        "acct_profile": acct_profile,
        "acct_data": acct_data,
        "status": Status.CREATED,
        "managed_state": managed_state,
        "invert_state": invert_state,
    }


async def validate(
    hub,
    name: str,
    sls_sources: List[str],
    render: str,
    runtime: str,
    subs: List[str],
    cache_dir: str,
    sls: List[str],
    target: str = None,
    test: bool = False,
    invert_state: bool = False,
    acct_file: str = None,
    acct_key: str = None,
    acct_profile: str = "default",
    acct_blob: bytes = None,
):
    """
    Run idem!
    """
    await hub.idem.state.create(
        name,
        sls_sources,
        render,
        runtime,
        subs,
        cache_dir,
        test,
        acct_file=acct_file,
        acct_key=acct_key,
        acct_profile=acct_profile,
        acct_blob=acct_blob,
        managed_state=None,
        param_sources=None,
        invert_state=invert_state,
    )
    # Get the sls file
    # render it
    # compile high data to "new" low data (bypass keyword issues)
    # Run the low data using act/idem
    hub.idem.RUNS[name]["params"] = hub.tool.parameter.get_validate_params()
    # process sls-sources
    hub.idem.RUNS[name]["status"] = Status.GATHERING
    await hub.idem.sls_source.init.gather(name, *sls)
    if hub.idem.RUNS[name]["errors"]:
        hub.idem.RUNS[name]["status"] = Status.GATHER_ERROR
        return
    if not hub.idem.RUNS[name].get("high"):
        # If there is no high data after rendering, that means the rendered result is empty. So we return.
        hub.idem.state.update_status(name, Status.FINISHED)
        return
    hub.idem.RUNS[name]["status"] = Status.COMPILING
    await hub.idem.state.compile(name)
    if hub.idem.RUNS[name]["errors"]:
        hub.idem.RUNS[name]["status"] = Status.COMPILATION_ERROR
        return
    await hub.idem.state.validate_(name)


async def apply(
    hub,
    name: str,
    sls_sources: List[str],
    render: str,
    runtime: str,
    subs: List[str],
    cache_dir: str,
    sls: List[str],
    target: str = None,
    test: bool = False,
    invert_state: bool = False,
    acct_file: str = None,
    acct_key: str = None,
    acct_profile: str = "default",
    acct_blob: bytes = None,
    acct_data: Dict[str, Any] = None,
    managed_state: Dict[str, Any] = None,
    param_sources: List[str] = None,
    params: List[str] = None,
):
    """
    Run idem!
    """
    if params is None:
        params = []
    await hub.idem.state.create(
        name,
        sls_sources,
        render,
        runtime,
        subs,
        cache_dir,
        test,
        acct_file=acct_file,
        acct_key=acct_key,
        acct_profile=acct_profile,
        acct_blob=acct_blob,
        acct_data=acct_data,
        managed_state=managed_state,
        param_sources=param_sources or [],
        invert_state=invert_state,
    )

    hub.idem.state.update_status(name, Status.GATHERING)
    if params:
        await hub.idem.sls_source.param.gather(name, *params)
        if hub.idem.RUNS[name]["errors"]:
            hub.idem.state.update_status(name, Status.GATHER_ERROR)
            return

    # Get the sls file
    # render it
    # compile high data to "new" low data (bypass keyword issues)
    # Run the low data using acct/idem
    await hub.idem.sls_source.init.gather(name, *sls)
    if hub.idem.RUNS[name]["errors"]:
        hub.idem.state.update_status(name, Status.GATHER_ERROR)
        return
    if not hub.idem.RUNS[name].get("high"):
        # If there is no high data after rendering, that means the rendered result is empty. So we return.
        hub.idem.state.update_status(name, Status.FINISHED)
        return
    hub.idem.state.update_status(name, Status.COMPILING)
    await hub.idem.state.compile(name)
    if hub.idem.RUNS[name]["errors"]:
        hub.idem.state.update_status(name, Status.COMPILATION_ERROR)
        return

    hub.idem.state.update_status(name, Status.RUNNING)
    await hub.idem.run.init.start(name, target=target)


async def compile_(hub, name):
    """
    Compile the data defined in the given run name
    """
    for mod in hub.idem.compiler:
        if hasattr(mod, "stage"):
            ret = mod.stage(name)
            await hub.pop.loop.unwrap(ret)


async def validate_(hub, name):
    for mod in hub.idem.validate:
        if hasattr(mod, "stage"):
            ret = mod.stage(name)
            await hub.pop.loop.unwrap(ret)


async def single(hub, _ref_: str, _test_: bool = None, *args, **kwargs):
    """
    Run a single state and return the raw result
    :param hub:
    :param _ref_: The state's reference on the hub
    :param _test_: Run the state in a low-consequence test-mode
    :param args: Args to be passed straight through to the state
    :param kwargs: Kwargs to be passed straight through to the state
    """
    if _test_ is None:
        _test_ = hub.OPT.idem.test

    acct_file = hub.OPT.acct.acct_file
    acct_key = hub.OPT.acct.acct_key
    acct_profile = hub.OPT.acct.get("acct_profile", "default")

    args = [a for a in args]

    if not _ref_.startswith("states."):
        _ref_ = f"states.{_ref_}"

    func = getattr(hub, _ref_)
    params = func.signature.parameters

    if "ctx" in params:
        ctx = await hub.idem.acct.ctx(
            path=_ref_,
            acct_file=acct_file,
            acct_key=acct_key,
            profile=acct_profile,
            hard_fail=True,
            validate=True,
        )
        ctx.test = _test_
        args.insert(0, ctx)

    ret = func(*args, **kwargs)
    return await hub.pop.loop.unwrap(ret)


async def batch(
    hub,
    states: Dict[str, Dict[str, Any]],
    runtime: str = None,
    test: bool = None,
    renderer: str = None,
    profiles: Dict[str, Any] = None,
    encrypted_profiles: str = None,
    acct_key: str = None,
    default_acct_profile: str = None,
    name: str = None,
    crypto_plugin: str = None,
    reconcile_plugin: str = None,
    pending_plugin: str = None,
    esm_plugin: str = None,
    esm_profile: str = None,
):
    """
    Run multiple states defined in code
    :param hub:
    :param states: A dictionary definition of the states to run
    :param runtime: "serial" or "parallel"
    :param test: Set "test" to "True" in the implicit ctx parameter
    :param renderer: The render pipe to use
    :param profiles: An unencrypted dump of acct profiles
    :param encrypted_profiles: An encoded dump of acct profiles encrypted with a crypto key
    :param acct_key: The decryption fernet key for acct profiles, defaults to the key defined in idem's runtime config
    :param default_acct_profile: The acct profile to use for states that don't have a profile explicitly defined
    :param name: A unique identifier for this batch's run
    :param crypto_plugin: The plugin that should be used to decrypt encrypted profile blobs
    :param reconcile_plugin: The plugin for reconciliation. Default: 'basic'
    :param pending_plugin: The plugin for determining state's 'pending' status for reconciliation
    :param esm_plugin: The plugin for esm
    :param esm_profile: The profile for esm
    """
    name = name or f"name_{uuid.uuid4()}"
    # Get these defaults from config
    crypto_plugin = crypto_plugin or hub.OPT.acct.crypto_plugin
    reconcile_plugin = reconcile_plugin or hub.OPT.idem.reconciler
    default_acct_profile = default_acct_profile or hub.OPT.idem.acct_profile
    runtime = runtime or hub.OPT.idem.runtime
    render = renderer or hub.OPT.idem.render
    test = hub.OPT.idem.test if test is None else test
    sls_source = f"sls_source_{uuid.uuid4()}"
    sls = [sls_source]
    data = {sls_source: states}
    sls_sources = [f"json://{json.dumps(data)}"]

    acct_key = acct_key or hub.OPT.acct.acct_key
    acct_profile = default_acct_profile or hub.OPT.acct.get("acct_profile", "default")

    reconcile_plugin = reconcile_plugin or hub.OPT.idem.get("reconciler", "basic")
    pending_plugin = pending_plugin or hub.OPT.idem.get("pending", "default")

    if profiles:
        if not acct_key:
            # If no acct_key exists, temporarily encrypt the profiles locally with a random key
            acct_key = hub.crypto[crypto_plugin].generate_key()
        elif encrypted_profiles:
            # combine the encrypted and unencrypted profiles into a single entity
            decrypted = hub.crypto[crypto_plugin].decrypt(encrypted_profiles, acct_key)
            profiles = dict_tools.update.update(profiles, decrypted, merge_lists=True)

        # Encrypt the raw profiles so that idem can easily consume them
        encrypted_profiles = hub.crypto[crypto_plugin].encrypt(profiles, acct_key)

    apply_kwargs = dict(
        name=name,
        sls_sources=sls_sources,
        render=render,
        runtime=runtime,
        subs=["states"],
        cache_dir=hub.OPT.idem.cache_dir,
        sls=sls,
        test=test,
        acct_key=acct_key,
        acct_profile=acct_profile,
    )
    if encrypted_profiles:
        if isinstance(encrypted_profiles, str):
            encrypted_profiles = encrypted_profiles.encode()
        apply_kwargs["acct_blob"] = encrypted_profiles
    else:
        # Default to the acct_file defined in the environment at runtime
        apply_kwargs["acct_file"] = hub.OPT.acct.acct_file

    context_manager = hub.idem.managed.context(
        run_name=name,
        cache_dir=hub.OPT.idem.cache_dir,
        esm_plugin=esm_plugin or hub.OPT.idem.esm_plugin,
        esm_profile=esm_profile or hub.OPT.idem.esm_profile,
        acct_key=acct_key,
        acct_blob=encrypted_profiles,
        serial_plugin=hub.OPT.idem.esm_serial_plugin,
    )
    try:
        async with context_manager as state:
            apply_kwargs["managed_state"] = state
            await hub.idem.state.apply(**apply_kwargs)

            if hub.idem.RUNS[name]["errors"]:
                # pop is required to remove the run from memory
                run_data = hub.idem.RUNS.pop(name)
                return run_data["errors"]

            # Reconciliation loop
            if test is False:
                await hub.reconcile.init.run(
                    plugin=reconcile_plugin,
                    pending_plugin=pending_plugin,
                    apply_kwargs=apply_kwargs,
                )
            else:
                # FINISHED event is triggered by the reconciliation.
                # Triggering the FINISHED event here since reconciliation is
                # skipped in 'test' mode.
                hub.idem.state.update_status(name, hub.idem.state.Status.FINISHED)
    except Exception as e:
        display = hub.output.nested.display(str(e))
        print(display)
        # Return a non-zero error code
        return 1

    run_data = hub.idem.RUNS.pop(name)
    return run_data["running"]


def status(hub, name: str) -> Dict[str, Any]:
    """
    Get the status of the named state run

        .. code-block:: json

            {
                "test": True or False,
                "errors": [],
                "running": {},
                "acct_profile": "acct profile name used for this run",
                "status": 0
                "status_name": "FINISHED"
            }
    :param hub:
    :param name: The unique identifier that was given to the state run
    :returns: A dictionary of the requested state run, An empty dictionary if the name doesn't exist
    """
    if name in hub.idem.RUNS:
        run = hub.idem.RUNS[name]
        hub.idem.state.update_status(name, run["status"])
        return {
            "test": run["test"],
            "errors": run["errors"],
            "running": run["running"],
            "acct_profile": run["acct_profile"],
            "status": run["status"].value,
            "status_name": run["status"].name,
        }
    else:
        hub.log.error(f"No idem run with Job ID: {name}")
        hub.idem.state.update_status(name, Status.UNDEFINED)
        return {
            "test": None,
            "errors": [],
            "running": {},
            "acct_profile": "",
            "status": Status.UNDEFINED.value,
            "status_name": Status.UNDEFINED.name,
        }



from typing import Any
from typing import Dict


def define(hub) -> Dict[str, Any]:
    """
    Return the definition used by the runtime to insert the conditions of the
    given requisite
    """
    return {
        "result": [True, None],
        "resolver": "any",
    }



"""
Process zip source
"""
import pathlib
import zipfile
from typing import ByteString
from typing import Tuple

__virtualname__ = "zip"

MIMETYPES = ["application/zip", "application/x-zip-compressed", "zip", ".zip"]


async def cache(
    hub, ctx, protocol: str, source: str, location: str
) -> Tuple[str, ByteString]:
    if zipfile.is_zipfile(source):
        zip_source = zipfile.ZipFile(source)

        try:
            # Normalize the path within the zipfile, it will always be a posix style path, even on windows
            location = pathlib.Path(location).as_posix()
            # This could be integrated with acct to read encrypted archives
            contents = zip_source.read(location, pwd=None)
            # Store the contents of the zip file in memory
            # Use <source>/<sls file> as the unique sls_ref
            return f"{source}/{location}", contents
        except KeyError:
            ...



import asyncio
import re
from typing import Any
from typing import Dict

import jmespath
import pop.contract
import pop.hub
import pop.loader


async def run(
    hub,
    desc_glob: str = "*",
    acct_file: str = None,
    acct_key: str = None,
    acct_profile: str = None,
    progress: bool = False,
    search_path: str = None,
    hard_fail: bool = False,
) -> Dict[str, Dict[str, Any]]:
    """
    :param hub:
    :param desc_glob:
    :param acct_file:
    :param acct_key:
    :param acct_profile:
    :param progress:
    :return:
    """
    result = {}
    coros = [
        _
        async for _ in hub.idem.describe.recurse(
            hub.states, desc_glob, acct_file, acct_key, acct_profile
        )
    ]

    if len(coros) == 0:
        # The current approach is to recurse through all loaded resources types.
        # If it reaches here, it should have gone through all possible resource types when looking up
        # Do not fail but instead log it, do not return from here as there might be 'auto_state' to process
        hub.log.warning(
            f"Cannot find anything to describe for resource type(s) '{desc_glob}'"
        )

    # This is going through dynamically generated states for exec modules
    # implementing __contracts__ = ["auto_state"]
    async for coro in hub.idem.describe.recurse(
        hub.exec, desc_glob, acct_file, acct_key, acct_profile
    ):
        coros.append(coro)

    if len(coros) == 0:
        # If there is none to process including 'auto_state', there is no progress to show. Return empty result
        return result

    # Create the progress bar
    progress_bar = hub.tool.progress.init.create(coros)

    for ret in asyncio.as_completed(coros):
        hub.tool.progress.init.update(progress_bar)
        try:
            result.update(await ret)
        except Exception as e:
            hub.log.error(f"Error during describe: {e.__class__.__name__}: {e}")
            if hard_fail:
                raise
    if progress:
        progress_bar.close()
    if search_path:
        prepped = hub.output.jmespath.prepare(result)
        searched = jmespath.search(search_path, prepped)
        if hub.OPT.rend.get("output") == "jmespath":
            # Don't post-process the result, it's already in jmespath format
            return searched
        else:
            return hub.output.jmespath.revert(searched)
    return result


async def recurse(
    hub,
    mod: pop.loader.LoadedMod or pop.hub.Sub,
    glob: str,
    acct_file: str,
    acct_key: str,
    acct_profile: str,
    ref: str = "",
):
    if hasattr(mod, "_subs"):
        for sub in mod._subs:
            if ref:
                r = ".".join([ref, sub])
            else:
                r = sub
            async for c in hub.idem.describe.recurse(
                mod[sub], glob, acct_file, acct_key, acct_profile, r
            ):
                yield c
    if hasattr(mod, "_loaded"):
        for loaded in mod._loaded:
            if ref:
                r = ".".join([ref, loaded])
            else:
                r = loaded
            async for c in hub.idem.describe.recurse(
                mod[loaded], glob, acct_file, acct_key, acct_profile, r
            ):
                yield c

    # Only describe functions that implement the "describe" contract
    if hasattr(mod, "__contracts__"):
        if "auto_state" in mod.__contracts__:
            # Handle auto state describe functions
            ctx = await hub.idem.acct.ctx(
                path=glob,
                acct_file=acct_file,
                acct_key=acct_key,
                profile=acct_profile,
                hard_fail=True,
                validate=True,
            )
            ctx.exec_mod_ref = ref
            coro = hub.states.auto_state.describe(ctx)
            yield hub.pop.Loop.create_task(coro)
        elif "resource" in mod.__contracts__:
            if re.fullmatch(glob, ref):
                ctx = await hub.idem.acct.ctx(
                    path=ref,
                    acct_file=acct_file,
                    acct_key=acct_key,
                    profile=acct_profile,
                    hard_fail=True,
                    validate=True,
                )
                if isinstance(mod.describe, pop.contract.ContractedAsync):
                    coro = mod.describe(ctx)
                else:

                    async def _async_desc():
                        return mod.describe(ctx)

                    coro = _async_desc()
                yield hub.pop.Loop.create_task(coro)
    elif (
        hasattr(mod, "_funcs") and "describe" in mod._funcs and re.fullmatch(glob, ref)
    ):
        raise ValueError(f"Implement the resource contract for '{ref}'.")



"""
This module is used to retrieve files from a json source
"""
import json
from typing import ByteString
from typing import Tuple

__virtualname__ = "json"


async def cache(
    hub, ctx, protocol: str, source: str, location: str
) -> Tuple[str, ByteString]:
    """
    Take a file from a location definition and cache it in memory
    """
    data = json.loads(source)

    key = hash(str(source))

    if location in data:
        zip_loc_data = data[location]
        data_cache = json.dumps(zip_loc_data).encode("utf-8")
        # use <source>/<sls file> as the unique sls_ref
        return f"{key}/{location}", data_cache



from typing import Any
from typing import Dict


def run(
    hub,
    seq: Dict[int, Dict[str, Any]],
    low: Dict[str, Any],
    running: Dict[str, Any],
    options: Dict[str, Any],
) -> Dict[int, Dict[str, Any]]:
    """
    Process the prereq system
    """
    pres = {}
    for ind, data in seq.items():
        if "prereq" not in data["chunk"]:
            continue
        chunk = data["chunk"]
        tag = data["tag"]
        pres[tag] = set()
        for rdef in chunk["prereq"]:
            state = next(iter(rdef))
            name = rdef[state]
            r_chunks = hub.idem.tools.get_chunks(low, state, name)
            for r_chunk in r_chunks:
                r_tag = hub.idem.tools.gen_chunk_func_tag(r_chunk)
                pres[tag].add(r_tag)
                reqret = {
                    "name": name,
                    "state": state,
                    "r_tag": r_tag,
                    "req": "prereq",
                    "chunk": r_chunk,
                    "ret": {},
                }
                data["reqrets"].append(reqret)
    if not pres:
        return seq
    unmet = {}
    for ind, data in seq.items():
        for tag, r_tags in pres.items():
            if tag not in unmet:
                unmet[tag] = set()
            if data["tag"] in r_tags:
                # Found one!
                if data["unmet"]:
                    unmet[tag].update(data["unmet"])
                else:
                    data["unmet"].add(tag)
    for ind, data in seq.items():
        tag = data["tag"]
        if tag in unmet:
            data["unmet"].update(unmet[tag])
    return seq



"""
The sls source is used to gather sls files, render them and return the initial
phase 1 highdata. This involves translating sls references into file paths,
downloading those sls files and then rendering them.

Once an sls file is rendered the include statements are resolved as well.
"""
import collections
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Set

__func_alias__ = {"iter_": "iter"}


def __init__(hub):
    hub.idem.resolve.init.HARD_FAIL = False


async def gather(hub, name: str, *sls, sources: List[str]) -> Dict[str, Any]:
    """
    Gather the named sls references

    :param hub:
    :param name: The state run name
    :param sls: sls locations within sources
    :param sources: sls-sources or params-sources
    """
    ret = {
        "state": {},
        "sls_refs": {},
        "blocks": {},
        "resolved": set(),
        "errors": [],
        "rendered": {},
        "order": [],
    }

    # Load sls content from sls files
    sls_content = collections.OrderedDict()
    sls_refs_map = {}
    for sls_relative_ref in sls:
        try:
            resolve_data = await hub.idem.get.ref(name, sls_relative_ref, sources)
            # TODO: This need to be discussed. The question is what to do when Idem sees the same sls file gets loaded
            #  again. The current behavior is, Idem doesn't fail, but it put this file to a later rendered order. See
            #  test_params_override_multiple_files test
            #  into a later order
            sls_ref = resolve_data["sls"]
            cfn = resolve_data["content"]

            if sls_ref in sls_content:
                sls_content.pop(sls_ref)
            sls_content[sls_ref] = cfn
            sls_refs_map[sls_ref] = sls_relative_ref
        except Exception as e:
            msg = f"Error while collecting blocks: {e.__class__.__name__}: {e}"
            hub.log.debug(msg)
            ret["errors"].append(msg)
            if hub.idem.resolve.init.HARD_FAIL:
                raise
            return ret

    # Render sls content into blocks
    block_data = await hub.idem.resolve.init.get_blocks(name, sls_content, sls_refs_map)
    ret.update(block_data)
    if ret["errors"]:
        return ret

    render_data = await hub.idem.resolve.init.render(
        name=name,
        blocks=block_data["blocks"],
        sls_refs=block_data["sls_refs"],
        resolved=block_data["resolved"],
    )
    unresolved_refs = render_data.pop("unresolved_refs", None)
    ret.update(render_data)

    if unresolved_refs:
        recurse = await hub.idem.resolve.init.gather(
            name, *unresolved_refs, sources=sources
        )
        ret["state"].update(recurse["state"])
        ret["rendered"].update(recurse["rendered"])
        ret["sls_refs"].update(recurse["sls_refs"])
        ret["blocks"].update(recurse["blocks"])
        ret["resolved"].update(recurse["resolved"])
        ret["errors"].extend(recurse["errors"])
        ret["order"].extend(recurse["order"])

    # Append the completely resolved sls files in order. This order is used for param override.
    ret["order"] += list(sls_content.keys())
    return ret


async def get_blocks(
    hub, name: str, sls_content: Dict, sls_ref_map: Dict
) -> Dict[str, Any]:
    """
    Returns the data from the SLS file if it can be found

    :param hub:
    :param name: The state run name
    :param sls_content: Unique sls_ref with sls file content in dict
    :param sls_ref_map: Map of the unique sls_ref to the relative sls file path within its source
    """
    ret = {"blocks": {}, "sls_refs": {}, "resolved": set(), "errors": []}
    for sls_ref, cfn in sls_content.items():
        if not cfn:
            msg = f"SLS ref '{sls_ref}' did not resolve"
            hub.log.debug(msg)
            ret["errors"].append(msg)
            return ret

        blocks = hub.rend.init.blocks(fn=sls_ref, content=cfn)
        if sls_ref in ret["sls_refs"]:
            msg = f"Discovered SLS ref collision between {ret['sls_refs']} and '{sls_ref}'. "
            hub.log.debug(msg)
            ret["errors"].append(msg)
            return ret
        ret["blocks"][sls_ref] = blocks
        ret["sls_refs"][sls_ref] = sls_ref_map[sls_ref]
        ret["resolved"].add(sls_ref)

    return ret


async def render(
    hub, name: str, blocks: Dict[str, Any], sls_refs: Dict[str, str], resolved: Set[str]
) -> Dict[str, Any]:
    """
    Pop the available blocks and render them if they have satisfied requisites

    :param hub:
    :param name: The state run name
    :param blocks: A mapping of sls refs to raw byte data representing a state
    :param sls_refs: References to sls within the given sources
    :param resolved: The names of SLSs that have been resolved
    """
    ret = {"rendered": {}, "state": {}, "unresolved_refs": []}
    for sls_ref, block in blocks.items():
        ret["state"][sls_ref] = {}
        cfn = sls_refs[sls_ref]
        for bname, chunk in block.items():
            clear = True
            for key, val in chunk.get("keys", {}).items():
                # If there is a render requisite plugin for this key then run its "check" method
                if key in hub.idem.resolve.requisite:
                    clear &= hub.idem.resolve.requisite[key].check(name, val)
                else:
                    clear = False

            if not clear:
                continue

            try:
                state = await hub.rend.init.parse_bytes(
                    chunk,
                    hub.idem.RUNS[name]["render"],
                    params=hub.idem.RUNS[name]["params"],
                )
            except Exception as e:
                msg = f"Error while parsing '{cfn}': {e}"
                hub.log.debug(msg)
                e.args = (msg,)
                raise e

            ret["rendered"][sls_ref] = bname
            if state is None:
                msg = f"SLS ref '{sls_ref}' is not resolved to any state."
                # if a rendered sls file turns to be empty
                hub.log.info(msg)
                continue
            # Process the state through the resolve plugins
            unresolved_refs = await hub.idem.resolve.init.apply(
                name=name,
                state=state,
                sls_ref=sls_ref,
                cfn=cfn,
                resolved=resolved,
            )
            ret["unresolved_refs"].extend(unresolved_refs)
            ret["state"][sls_ref].update(state)

    for sls_ref, bname in ret["rendered"].items():
        blocks[sls_ref].pop(bname, None)
    return ret


async def apply(
    hub,
    name: str,
    state: Dict[str, Any],
    sls_ref: str,
    cfn: str,
    resolved: Set[str],
) -> List[str]:
    """
    Introduce the raw state into the running dataset

    :param hub:
    :param name: The state run name
    :param state: A rendered block from the sls
    :param sls_ref: A reference to another sls within the given sources
    :param cfn: The cache file name, or the location of sls within the given sources
    :param resolved: a set of refs that have already been resolved
    """
    unresolved_refs = []
    if not isinstance(state, Dict):
        hub.idem.RUNS[name]["errors"].append(
            f"SLS {sls_ref} is not formed as a dict but as a {type(state)}"
        )
        return unresolved_refs

    for top_level_resolver in hub.idem.resolve._loaded:
        if top_level_resolver == "init":
            # We are currently in resolve.init.apply, avoid infinite recursion
            continue
        ret = await hub.idem.resolve[top_level_resolver].apply(
            name=name,
            state=state,
            sls_ref=sls_ref,
            cfn=cfn,
            resolved=resolved,
        )
        unresolved_refs.extend(ret)
    return unresolved_refs


def iter_(hub, state: Dict[str, Any]) -> Generator:
    """
    iterate over a state, skipping known keywords

    :param hub:
    :param state: A rendered block from the sls
    """
    for id_ in state:
        if any(id_ in getattr(plugin, "KEYWORDS", []) for plugin in hub.idem.resolve):
            continue
        yield id_



def check(hub, name: str, value: str) -> bool:
    """
    Verify that states using a "require" shebang have had their requisite met

    :param hub:
    :param name: The state run name
    :param value: A string value to find in running states

    .. code-block:: yaml

        state_1:
          test.nop

        #!require:state_1
        state_2:
          test.nop:
    """
    for tag, data in hub.idem.RUNS[name]["running"].items():
        if data["name"] == value:
            return True
    return False
